#include "mainwindow.h"
#include <QMenu>
#include <QMenuBar>
#include <QMessageBox>
#include <QCloseEvent>
#include <QToolBar>
#include <QSpinBox>
#include <QIcon>
#include <QFileDialog>
#include <QImage>
#include <QFile>
#include <QTextStream>
#include <QComboBox>
#include <QtDebug>

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent),hasChanged(false),inSaveAs(false)
{
    statusbar = new QStatusBar(this);
    setStatusBar(statusbar);
    setCentralWidget(view = new QFraktalView(this));
    initActions();
    initToolBar();
    QFile file(":/help/help_en.html");
    file.open(QIODevice::ReadOnly);
    QTextStream txt(&file);
    helptext = txt.readAll();
    connect(view,SIGNAL(signalStatusChanged(QString)),statusbar,SLOT(showMessage(QString)));
    connect(view,SIGNAL(newImageRendered()),this,SLOT(slotViewContentChanged()));
}

MainWindow::~MainWindow()
{

}

void MainWindow::initActions()
{
    QMenuBar *menubar=new QMenuBar(this);
    QMenu *fileMenu = new QMenu(tr("&File"));
    fileNew = fileMenu->addAction(QIcon(":/icons/filenew.png"),tr("&New"),this,SLOT(slotFileNew()),QKeySequence::New);
    fileOpen = fileMenu->addAction(QIcon(":/icons/fileopen.png"),tr("&Open"),this,SLOT(slotFileOpen()),QKeySequence::Open);
    fileSave = fileMenu->addAction(QIcon(":/icons/filesave.png"),tr("&Save"),this,SLOT(slotFileSave()),QKeySequence::Save);
    fileSaveAs = fileMenu->addAction(QIcon(":/icons/filesaveas.png"),tr("Save &as"),this,SLOT(slotFileSaveAs()),QKeySequence::SaveAs);
    fileSnapshot = fileMenu->addAction(QIcon(":/icons/snapshot.png"),tr("Save image"),this,SLOT(slotFileSnapshot()),QKeySequence::Print);
    QMenu *threadMenu = new QMenu(tr("Threads"),this);
    threadMenu->addActions(view->threadNumActions->actions());
    fileMenu->addMenu(threadMenu);
    fileQuit = fileMenu->addAction(QIcon(":/icons/exit.png"),tr("&Quit"),this,SLOT(slotFileQuit()),QKeySequence::Close);
    menubar->addMenu(fileMenu);
    QMenu *editMenu = new QMenu(tr("&Edit"));
    editCopy = editMenu->addAction(tr("&Copy"),this,SLOT(slotEditCopy()),QKeySequence::Copy);
    menubar->addMenu(editMenu);
    QMenu *viewMenu = new QMenu(tr("&View"));
    viewMenu->addAction(view->zoomIn);
    viewMenu->addAction(view->zoomOut);
    viewMenu->addAction(view->resetZoom);
    QMenu *viewColorMapMenu=new QMenu(tr("Colormaps"));
    viewColorMapMenu->addActions(view->colormapActions->actions());
    viewMenu->addMenu(viewColorMapMenu);
    viewMenu->addAction(view->antialias);
    menubar->addMenu(viewMenu);
    QMenu *helpMenu = new QMenu(tr("&Help"));
    helpMenu->addAction(tr("Help..."),this,SLOT(slotHelpHelp()),QKeySequence::HelpContents);
    helpMenu->addAction(tr("About..."),this,SLOT(slotHelpAbout()));
    helpMenu->addAction(tr("About Qt"),this,SLOT(slotHelpAboutQt()));
    menubar->addMenu(helpMenu);
    this->setMenuBar(menubar);
}

void MainWindow::initToolBar()
{
    QToolBar *filebar = new QToolBar(tr("File operations"),this);
    fileNew->setToolTip(tr("Creates a new program instance."));
    filebar->addAction(fileNew);
    fileOpen->setToolTip(tr("Loads a mandelbrot set from file."));
    filebar->addAction(fileOpen);
    fileSave->setToolTip(tr("Saves a mandelbrot set to file."));
    filebar->addAction(fileSave);
    fileSaveAs->setToolTip(tr("Saves the mandelbrot set under a new name."));
    filebar->addAction(fileSaveAs);
    fileSnapshot->setToolTip(tr("Saves the mandelbrot image."));
    filebar->addAction(fileSnapshot);
    fileQuit->setToolTip(tr("Quits the application instance."));
    filebar->addAction(fileQuit);
    this->addToolBar(filebar);
    QToolBar *viewbar = new QToolBar(tr("View settings"),this);
    iteration_spinner = new QSpinBox(this);
    iteration_spinner->setMinimum(10);
    iteration_spinner->setMaximum(10000);
    iteration_spinner->setValue(view->getCurrentIterationDepth());
    viewbar->addWidget(iteration_spinner);
    connect(iteration_spinner,SIGNAL(valueChanged(int)),view,SLOT(slotSetIterationDepth(int)));
    iteration_spinner->setToolTip(tr("Sets the iteration depth for the mandelbrot set."));
    viewbar->addAction(view->zoomIn);
    viewbar->addAction(view->zoomOut);
    viewbar->addAction(view->resetZoom);
    viewbar->addAction(view->antialias);
    this->addToolBar(viewbar);
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    if(hasChanged)
    {
        switch(QMessageBox::question(this,"QFraktal",tr("Unsaved changes! Quit anyway?"),QMessageBox::Ok | QMessageBox::Save | QMessageBox::Abort,QMessageBox::Save))
        {
            case QMessageBox::Abort: event->ignore();
                break;
            case QMessageBox::Save:
                hasChanged=false;
                slotFileSave();
                break;
            default: break;
        };

    }
    else this->close();
}

void MainWindow::slotFileSave()
{
    if(lastFileName.isNull())
    {
        slotFileSaveAs();
    }
    else
    {
        QFile file(lastFileName);
        try
        {
            if(file.exists() && inSaveAs)
            {
                switch(QMessageBox::warning(this,tr("QFraktal warning"),tr("File %1 still exists. Rename?").arg(lastFileName),QMessageBox::Abort | QMessageBox::Ok | QMessageBox::Save,QMessageBox::Ok))
                {
                    case QMessageBox::Abort: throw tr("Saving aborted.");
                        break;
                    case QMessageBox::Ok:
                        lastFileName = QFileDialog::getSaveFileName(this,tr("QFraktal save fraktal"),"/home",tr("XML files (*.xml);;All files (*)"));
                        break;
                    default:
                        break;
                };
            }
            if(!file.open(QIODevice::WriteOnly)) throw tr("Could not open file %1 for writing!").arg(lastFileName);
            QDomDocument doc(doctype);
            doc.createComment(tr("QFraktal save file, QFraktal by R&uuml;diger Kn&ouml;rig"));
            QDomElement root=doc.createElement("Mandelbrot");
            doc.appendChild(root);
            QDomElement itdepth = doc.createElement("iteration_depth");
            root.appendChild(itdepth);
            itdepth.appendChild(doc.createTextNode(QString("%1").arg(view->getCurrentIterationDepth())));
            QDomElement colormap = doc.createElement("colormap");
            colormap.appendChild(doc.createTextNode(view->colormapActions->checkedAction()->text()));
            root.appendChild(colormap);
            QDomElement area = doc.createElement("viewed_area");
            root.appendChild(area);
            QDomElement upperleft = doc.createElement("upperleft");
            upperleft.setAttribute("real",view->getUpperLeftCorner().real());
            upperleft.setAttribute("imag",view->getUpperLeftCorner().imag());
            area.appendChild(upperleft);
            QDomElement lowerright = doc.createElement("lowerright");
            lowerright.setAttribute("real",view->getLowerRightCorner().real());
            lowerright.setAttribute("imag",view->getLowerRightCorner().imag());
            area.appendChild(lowerright);
            QTextStream os(&file);
            os << doc.toString(1);
            if(file.error() != QFile::NoError) throw tr("Could not write to file %1!").arg(lastFileName);
            file.close();
            hasChanged=false;
        }
        catch(QString msg)
        {
            msg=tr("%1\nReported Reason:%2").arg(msg).arg(file.errorString());
            QMessageBox::warning(this,tr("QFraktal save error"),msg);
        }
    }
}

void MainWindow::slotFileSaveAs()
{
    inSaveAs=true;
    lastFileName = QFileDialog::getSaveFileName(this,tr("QFraktal save fraktal"),"/home",tr("XML files (*.xml);;All files (*)"));
    slotFileSave();
    inSaveAs=false;
}

void MainWindow::slotFileSnapshot()
{
    QString filename=QFileDialog::getSaveFileName(this,tr("QFractal save image."),"/home",tr("Images (*.jpg *.png *.xpm);; All files (*)"));
    if(!filename.isNull())
    {
        if(!view->getRenderedImage().save(filename))
        {
            QMessageBox::warning(this,tr("QFraktal save error."),tr("Could not write image file %1. Forgot file suffix (e.g. .jpg)?").arg(filename));
        }
    }
}

void MainWindow::slotFileOpen()
{
    QString fileName = QFileDialog::getOpenFileName(this,tr("QFraktal open a file."),"/home",tr("XML files (*.xml);;All files (*)"));
    QFile file(fileName);
    if(!fileName.isNull())
    {
        try{
            if(!file.open(QIODevice::ReadOnly)) throw tr("Could not open file %1!\nReported Reason:%2").arg(fileName).arg(file.errorString());
             QDomDocument doc(doctype);
             if(!doc.setContent(&file)) throw tr("Not a valid XML file!");
             if(doc.doctype().name() != doctype) throw tr("Not a QFraktal settings file!");
             QDomElement it = lookupElement(doc,"iteration_depth");
             bool ok;
             int iteration_depth=it.text().toInt(&ok,10);
             if(!ok) throw tr("Error while parsing the iteration_depth element!");
             QDomElement ul_e = lookupElement(doc,"upperleft");
             complex<double> ul;
             ul.real() = ul_e.attribute("real").toDouble(&ok);
             if(!ok) throw tr("Could not parse real part of the upperleft element!");
             ul.imag() = ul_e.attribute("imag").toDouble(&ok);
             if(!ok) throw tr("Could not parse imaginary part of the upperleft element!");
             QDomElement lr_e = lookupElement(doc,"lowerright");
             complex<double> lr;
             lr.real() = lr_e.attribute("real").toDouble(&ok);
             if(!ok) throw tr("Could not parse real part of the lowerright element!");
             lr.imag() = lr_e.attribute("imag").toDouble(&ok);
             if(!ok) throw tr("Could not parse imaginary part of the lowerright element!");
             QString colormapname = lookupElement(doc,"colormap").text();
             QList<QAction *> colormaplist = view->colormapActions->actions();
             QList<QAction *>::iterator li=colormaplist.begin();
             for(;li != colormaplist.end();li++) if((*li)->text()==colormapname) break;
             if(li == colormaplist.end())
             {
                QMessageBox::warning(this,"QFraktal",tr("Colormap %1 not available!").arg(colormapname));
             }
             else (*li)->setChecked(true);

             view->slotSetView(ul,lr,false);
             if(iteration_depth==iteration_spinner->value())
             {
                 view->slotSetIterationDepth(iteration_depth);
             }
             else
             {
                 iteration_spinner->setValue(iteration_depth);
             }
             hasChanged=false;
        }catch(QString msg)
        {
            QMessageBox::warning(this,tr("QFraktal load error"),msg);
        }
    }
}

const QString MainWindow::about=tr("<html><body><h1>QFraktal</h1><p>Mandelbrot set viewer programmed by R&uuml;diger Kn&ouml;rig in Saigon September 2009.</p></body></html>");

//const QString MainWindow::helptext=tr("<html></html>");

const QString MainWindow::doctype="de.knoerig.www.QFraktalSettings.dtd";
