#ifndef QFRAKTALVIEW_H
#define QFRAKTALVIEW_H

#include <QWidget>
#include <QThread>
#include <QString>
#include <QColor>
#include <QImage>
#include <QPixmap>
#include <QPoint>
#include <QMouseEvent>
#include <QPainter>
#include <QResizeEvent>
#include <QPaintEvent>
#include <QContextMenuEvent>
#include <QAction>
#include <QMenu>
#include <QActionGroup>
#include <QTime>
#include <complex>
#include <vector>
#include "QColorMapAction.h"
#include <QMutex>
#include "renderthread.h"
#include "RenderNumAction.h"

using namespace std;

class QFraktalView : public QWidget
{
        Q_OBJECT
public:
    /**
     * @brief Constructor.
     */
    QFraktalView(QWidget *parent) : QWidget(parent),iteration_depth(100),isRendering(false),usedRenderers(1)
    {
        setCursor(Qt::CrossCursor);
        for(int i=0;i<numThreads;i++)
        {
            connect(&renderer[i],SIGNAL(renderingCompleted(const QImage &,const QRect &)),this,SLOT(slotRenderingCompleted(const QImage &,const QRect &)));
        }
        initActions();
        setMouseTracking(true);
        setIterationDepth(100);
        slotResetView();
    }

    ~QFraktalView()  { }

    /**
     * @brief Gives the window system a hint of the desired widget size.
     */
    QSize sizeHint() const { return QSize(400,400); }

    /**
     * @brief Getter for a const reference to the rendered image.
     * If you get a null image, none has been calculated so far.
     * @return Const reference to the rendered image.
     */
    const QImage &getRenderedImage() const { return fraktal; }

    /**
     * @brief Getter for the current iteration depth.
     * @return Current iteration depth.
     */
    int getCurrentIterationDepth() const { return iteration_depth; }

    /**
     * @brief Inspector for the upper-left corner of the viewed area in the complex plane.
     * @return Upper-left corner of the viewed area in the complex plane.
     */
    const complex<double> &getUpperLeftCorner() const { return ul; }

    /**
     * @brief Inspector for the lower-right corner of the viewed area in the complex plane.
     * @return Lower-right corner of the viewed area in the complex plane.
     */
    const complex<double> &getLowerRightCorner() const { return lr; }

    /**
     * @brief Setup method.
     * Setup all parameters with only one rerendering.
     * @param ul Upper-left corner of the viewed area.
     * @param lr Lower-right corner of the viewed area.
     * @param itdepth new value for the iteration depth.
     */
    void setup(complex<double> ul, complex<double> lr, int itdepth)
    {
        setIterationDepth(itdepth);
        slotSetView(ul,lr);
    }

public slots:
    /**
     * @brief Reset the viewed area of the complex plane to default values.
     */
    void slotResetView()
    {
        slotSetView(complex<double>(-1.8,1.2),complex<double>(0.65,-1.2));
    }

    /**
     * @brief Slot for setting the viewed area.
     * @param ul Upper-left corner of the viewed area.
     * @param lr Lower-right corner of the viewed area.
     * @param rerender Rerendering yes or no?
     */
    void slotSetView(complex<double> ul,complex<double> lr,bool rerender=true)
    {
        setView(ul,lr);
        if(rerender) renderFractal();
    }

    /**
     * @brief Slot for setting the iteration depth and trigger rerendering.
     * @param itdepth new value for the iteration depth.
     */
    void slotSetIterationDepth(int itdepth)
    {
        setIterationDepth(itdepth);
        renderFractal();
    }

    void slotZoomIn()
    {
        complex<double> v=(lr-ul)*0.1;
        slotSetView(ul+v,lr-v);
    }

    void slotZoomOut()
    {
        complex<double> v=(lr-ul)*0.1;
        slotSetView(ul-v,lr+v);
    }

public: // public attributes
    QAction *resetZoom; /**< Resets the viewed area to default */
    QAction *zoomIn; /**< Zooms in */
    QAction *zoomOut; /**< Zooms out */
    QActionGroup *colormapActions; /**< Actions to provide the color maps */
    QActionGroup *threadNumActions; /**< Actions to change the number of rendering threads */
    QAction *antialias; /**< toggles the anti-alias rendering on or off */

signals:
    /**
     * @brief Emitted when a status bar change should occur.
     */
    void signalStatusChanged(QString msg);

    /**
     * @brief Emitted when a new image has been rendered.
     */
    void newImageRendered();

protected: // protected methods

    /**
     * @brief For setting the viewed area without rendering.
     * @param ul Upper-left corner of the viewed area.
     * @param lr Lower-right corner of the viewed area.
     */
    void setView(complex<double> ul,complex<double> lr)
    {
        this->ul=ul;
        this->lr=lr;
        recalculateStepSizes();
        emit signalStatusChanged(tr("View area from (%1,%2) to (%3,%4).").arg(ul.real()).arg(ul.imag()).arg(lr.real()).arg(lr.imag()));
        emit newImageRendered();
    }

     /**
     * @brief For setting the iteration depth without trigger rerendering.
     * @param itdepth new value for the iteration depth.
     */
    void setIterationDepth(int itdepth)
    {
        iteration_depth=itdepth;
        colormap.resize(iteration_depth);
        static_cast<QColorMapAction *>(colormapActions->checkedAction())->apply(colormap);
    }
    /**
     * @brief Reimplementation of QWidget::resizeEvent(QResizeEvent *).
     * @see QWidget::resizeEvent(QResizeEvent *)
     */
    void resizeEvent(QResizeEvent *e)
    {
        fraktal = QImage(e->size(),QImage::Format_ARGB32);
        recalculateStepSizes();
        renderFractal();
    }

    /**
     * @brief Recalculates the step sizes dr and di.
     * @see dr
     * @see di
     */
    void recalculateStepSizes()
    {
        complex<double> diff=lr-ul;
        dr=diff.real()/width();
        di=diff.imag()/height();
    }


    /**
     * @brief Reimplementation of QWidget::paintEvent(QPaintEvent *).
     * @see QWidget::paintEvent(QPaintEvent *)
     */
    void paintEvent(QPaintEvent *e)
    {
        if(backbuffer.size()==size())
        {
            QPainter p(this);
            p.drawPixmap(e->rect(),backbuffer,e->rect());
            if(!selection.isNull())
            {
                p.setPen(QPen(QColor(255,255,255)));
                p.drawRect(selection);
            }
        }
    }

    void mouseMoveEvent(QMouseEvent *);

    void mousePressEvent(QMouseEvent *);

    void mouseReleaseEvent(QMouseEvent *);

    void contextMenuEvent(QContextMenuEvent *e) { contextmenu->popup(e->globalPos()); }


    /**
     * @brief Inits the actions used here.
     */
    void initActions();

protected slots:

    /**
     * @brief Renders the fractal.
     */
    void renderFractal();

    /**
     * @brief Slot to change the colormap.
     * @param action Colormap action casted to QAction.
     */
    void slotColormapChanged(QAction *action)
    {
        static_cast<QColorMapAction *>(action)->apply(colormap);
        renderFractal();
    }

    /**
     * @brief Gets called when a renderer completes his task.
     * @param tile rendered image tile
     * @param region region of the image tile
     */
    void slotRenderingCompleted(const QImage &tile,const QRect &region)
    {
        QPainter p(&fraktal);
        p.drawImage(region.topLeft(),tile);
        if(--activeRenderers <= 0)
        {
            emit signalStatusChanged(tr("Rendering time: %1 ms").arg(stopwatch.elapsed()));
            backbuffer = QPixmap::fromImage(fraktal);
            isRendering=false;
            update();
            setCursor(Qt::CrossCursor);
            threadNumActions->setEnabled(true);
            antialias->setEnabled(true);
        }
    }

    void slotSetNumThreads() { usedRenderers=((RenderNumAction *) sender())->getNumberOfThreads(); renderFractal(); }

protected: // protected attributes
    complex<double> ul; /**< upper-left corner in the complex plane */
    complex<double> lr; /**< lower-right corner in the complex plane */
    double dr; /**< step size for the real part */
    double di; /**< step size for the imaginary part */
    int iteration_depth; /**< current iteration depth */
    vector<QRgb> colormap; /**< colormap for mapping iteration depth to colors */
    QImage fraktal; /**< rendered fractal */
    QPixmap backbuffer; /**< backbuffer copy of the rendered image */
    bool isRendering; /**< flag indicating if rendering takes place. */
    QRect selection;/**< stores the position of the zoom selection */
    QMenu *contextmenu; /**< Right-click context menu */
    enum { numThreads=4 }; /**< number of available processing threads */
    RenderThread renderer[numThreads]; /**< the four rendering threads available here */
    int usedRenderers; /**< number of actually used */
    int activeRenderers; /**< number of processing renderers */
    QTime stopwatch; /**< for measuring rendering times */
};

#endif // QFRAKTALVIEW_H
