#include "renderthread.h"
#include <QMutexLocker>

RenderThread::~RenderThread()
{
    mutex.lock();
    abort = true;
    condition.wakeOne();
    mutex.unlock();
    // wait that thread execution finished
    wait();
}

void RenderThread::render(complex<double> ul, complex<double> lr,double dr,double di,const vector<QRgb> colormap,QRect region,bool antialias)
{
    // get unique access to member data
    QMutexLocker locker(&mutex);
    this->ul=ul;
    this->lr=lr;
    this->dr=dr;
    this->di=di;
    this->region=region;
    this->colormap=colormap;
    this->antialias=antialias;

    if (!isRunning()) {
         start(LowPriority);
     } else {
         restart = true;
         condition.wakeOne();
     }
}

void RenderThread::run()
{
    forever
    {
        mutex.lock();
        complex<double> dv=lr-ul;
        int iteration_depth=colormap.size();
        int ymin=region.top();
        //int ymax=region.bottom();
        int xmin=region.left();
        //int xmax=region.right();
        int width=region.width();
        int height=region.height();
        double rl=ul.real()+dr*xmin;
        complex<double> c=complex<double>(rl,ul.imag()+di*ymin);
        if(antialias) { width<<=1; height<<=1; dr*=0.5; di*=0.5; }
        QImage image(QSize(width,height),QImage::Format_ARGB32);
        mutex.unlock();

        for(int y=0;y<height;y++)
        {
            if(restart) break;
            if(abort)return;
            QRgb *ptr=reinterpret_cast<QRgb *>(image.scanLine(y));
            for(int x=0;x<width;x++)
            {
                complex<double> z=complex<double>(0,0);
                int it=0;
                for(;it<iteration_depth;it++)
                {
                    z=z*z+c;
                    if(z.real()*z.real()+z.imag()*z.imag()>4.0) break;
                }
                *(ptr++)=((it < iteration_depth) ? colormap[it] : qRgb(0,0,0));
                c.real()+=dr;
            }
            c.imag()+=di;
            c.real()=rl;
        }
        if(antialias) image=image.scaled(width>>1,height>>1,Qt::IgnoreAspectRatio,Qt::SmoothTransformation);
        if(!restart) emit renderingCompleted(image,region);
        mutex.lock();
        if (!restart) condition.wait(&mutex);
        restart = false;
        mutex.unlock();
    }
}
